import functools
import json

from cms.utils.plugins import get_bound_plugins
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder

from . import custom_process_hook, get_serializer_name
from .utils import get_plugin_fields

dump_json = functools.partial(json.dumps, cls=DjangoJSONEncoder)


def get_plugin_data(plugin, only_meta=False):
    if only_meta:
        custom_data = None
    else:
        plugin_fields = get_plugin_fields(plugin.plugin_type)
        _plugin_data = serializers.serialize(
            get_serializer_name(), (plugin,), fields=plugin_fields
        )[0]
        custom_data = _plugin_data["fields"]

    plugin_data = {
        "pk": plugin.pk,
        "creation_date": plugin.creation_date,
        "position": plugin.position,
        "plugin_type": plugin.plugin_type,
        "parent_id": plugin.parent_id,
        "data": custom_data,
    }

    # customize plugin-data on export
    return custom_process_hook(
        "DJANGOCMS_TRANSFER_PROCESS_EXPORT_PLUGIN_DATA",
        plugin,
        plugin_data
    )


def export_plugin(plugin):
    data = get_plugin_export_data(plugin)
    return dump_json(data)


def export_placeholder(placeholder, language):
    data = get_placeholder_export_data(placeholder, language)
    return dump_json(data)


def export_page(cms_pagecontent, language):
    data = get_page_export_data(cms_pagecontent, language)
    return dump_json(data)


def get_plugin_export_data(plugin):
    descendants = plugin.get_descendants()
    plugin_data = [get_plugin_data(plugin=plugin)]
    plugin_data[0]["parent_id"] = None
    plugin_data.extend(
        get_plugin_data(plugin) for plugin in get_bound_plugins(descendants)
    )
    return plugin_data


def get_placeholder_export_data(placeholder, language):
    plugins = placeholder.get_plugins(language)
    # The following results in two queries;
    # First all the root plugins are fetched, then all child plugins.
    # This is needed to account for plugin path corruptions.

    return [get_plugin_data(plugin) for plugin in get_bound_plugins(list(plugins))]


def get_page_export_data(cms_pagecontent, language):
    data = []
    placeholders = cms_pagecontent.rescan_placeholders().values()

    for placeholder in list(placeholders):
        plugins = get_placeholder_export_data(placeholder, language)
        data.append({"placeholder": placeholder.slot, "plugins": plugins})
    return data



import pytest

from djangocms_transfer import custom_process_hook

from .abstract import FunctionalityBaseTestCase


class CustomProcessHookTest(FunctionalityBaseTestCase):
    def setUp(self):
        super().setUp()
        self.plugin = self._create_plugin()
        self.plugin_data = self._get_expected_page_export_data

    def test_empty_nonexistent_custom_process_hook(self):
        self.assertEqual(
            custom_process_hook("", self.plugin, self.plugin_data),
            self.plugin_data
        )
        self.assertEqual(
            custom_process_hook("UNKNOWN_TRANSFER", self.plugin, self.plugin_data),
            self.plugin_data
        )

    @pytest.mark.usefixtures("use_nonexistent_transfer_hook")
    def test_nonexistent_module(self):
        self.assertRaises(
            ImportError,
            custom_process_hook,
            "DJANGOCMS_TRANSFER_PROCESS_IMPORT_PLUGIN_DATA",
            self.plugin, self.plugin_data
        )
        self.assertRaises(
            ImportError,
            custom_process_hook,
            "DJANGOCMS_TRANSFER_PROCESS_EXPORT_PLUGIN_DATA",
            self.plugin, self.plugin_data
        )

    @pytest.mark.usefixtures("use_existent_transfer_hook")
    def test_existent_module(self):
        self.assertTrue(
            custom_process_hook(
                "DJANGOCMS_TRANSFER_PROCESS_IMPORT_PLUGIN_DATA",
                self.plugin, {}
            ),
            (self.plugin, {})
        )
        self.assertTrue(
            custom_process_hook(
                "DJANGOCMS_TRANSFER_PROCESS_EXPORT_PLUGIN_DATA",
                self.plugin, self.plugin_data
            ),
            (self.plugin, self.plugin_data)
        )



import json

from cms.plugin_pool import plugin_pool
from cms.utils.urlutils import admin_reverse
from django.core.exceptions import PermissionDenied
from django.core.files.uploadedfile import SimpleUploadedFile

from .abstract import FunctionalityBaseTestCase


class PluginImporterTestCase(FunctionalityBaseTestCase):
    def setUp(self):
        super().setUp()
        self.user = self._create_user("test", True, True)
        self.plugin_importer = next(
            (
                plugin
                for plugin in plugin_pool.get_all_plugins()
                if plugin.__name__ == "PluginImporter"
            ),
            None
        )

    def test_get_plugin_urls(self):
        urlpatterns = self.plugin_importer().get_plugin_urls()
        self.assertEqual(len(urlpatterns), 2)
        self.assertEqual(urlpatterns[0].name, "cms_export_plugins")
        self.assertEqual(urlpatterns[1].name, "cms_import_plugins")

    def test_get_extra_menu_items(self):
        request = self.get_request()

        with self.subTest("extra plugin menu items"):
            text_plugin = self._create_plugin()
            pluginimporter_plugin = self._add_plugin_to_page("PluginImporter")
            menu_items = self.plugin_importer.get_extra_plugin_menu_items(request, text_plugin)
            self.assertEqual(len(menu_items), 2)
            self.assertEqual(menu_items[0].name, "Export plugins")
            self.assertEqual(menu_items[1].name, "Import plugins")
            self.assertEqual(
                menu_items[0].url,
                "/en/admin/cms/page/plugin/plugin_importer/export-plugins/?language=en&plugin=1"
            )
            self.assertEqual(
                menu_items[1].url,
                "/en/admin/cms/page/plugin/plugin_importer/import-plugins/?language=en&plugin=1"
            )
            # no menu item for PluginImporter itself
            self.assertIsNone(self.plugin_importer.get_extra_plugin_menu_items(request, pluginimporter_plugin))

        with self.subTest("extra placeholder menu items"):
            placeholder = self.page_content.get_placeholders().get(slot="content")
            menu_items = self.plugin_importer.get_extra_placeholder_menu_items(request, placeholder)
            self.assertEqual(
                menu_items[0].url,
                "/en/admin/cms/page/plugin/plugin_importer/export-plugins/?language=en&placeholder=1"
            )
            self.assertEqual(
                menu_items[1].url,
                "/en/admin/cms/page/plugin/plugin_importer/import-plugins/?language=en&placeholder=1"
            )

    def test_import_plugin_views(self):
        with self.login_user_context(self.user):
            response = self.client.get(admin_reverse("cms_import_plugins"))
            self.assertEqual(response.content, b"Form received unexpected values.")

        del self.user # self.get_request() checks for "user" attribute
        self.assertRaises(
            PermissionDenied,
            self.plugin_importer.import_plugins_view,
            self.get_request()
        )

    def test_import_plugin_views_for_page(self):
        with self.login_user_context(self.user):
            # GET page import
            response = self.client.get(
                admin_reverse("cms_import_plugins") + f"?language=en&cms_pagecontent={self.page_content.id}"
            )
            self.assertEqual(response.templates[0].name, "djangocms_transfer/import_plugins.html")
            self.assertEqual(response.context["form"].initial["cms_pagecontent"], self.page_content)

            # POST page import
            post_data = {
                "language": "en", "cms_pagecontent": [self.page_content.id],
                "import_file": SimpleUploadedFile("file.txt", bytes(json.dumps(self._get_expected_page_export_data()), "utf-8"))
            }
            response = self.client.post(
                admin_reverse("cms_import_plugins") + f"?language=en&cms_pagecontent={self.page_content.id}",
                post_data
            )
            self.assertIn(b'<div class="success"></div>', response.content)

    def test_import_plugin_views_for_placeholder(self):
        placeholder = self.page_content.get_placeholders().get(slot="content")
        with self.login_user_context(self.user):
            # GET placeholder import
            response = self.client.get(
                admin_reverse("cms_import_plugins") + f"?language=en&placeholder={placeholder.id}"
            )
            self.assertEqual(response.templates[0].name, "djangocms_transfer/import_plugins.html")
            self.assertEqual(response.context["form"].initial["placeholder"], placeholder)

            # empty placeholder import (no existing plugin)
            request_path = admin_reverse("cms_import_plugins") + f"?language=en&placeholder={placeholder.id}"
            post_data = {
                "language": "en", "placeholder": [placeholder.id],
                "import_file": SimpleUploadedFile("file.txt", b"[null, null]")
            }
            with self.assertRaises(IndexError):
                self.client.post(path=request_path, data=post_data)

            # create plugins in the placeholder
            text_plugin = self._create_plugin()
            self._add_plugin_to_page("PluginImporter")
            self._create_plugin(plugin_type="LinkPlugin", parent=text_plugin)

            # empty placeholder import
            request_path = admin_reverse("cms_import_plugins") + f"?language=en&placeholder={placeholder.id}"
            post_data = {
                "language": "en", "placeholder": [placeholder.id],
                "import_file": SimpleUploadedFile("file.txt", b"[null, null]")
            }
            response = self.client.post(path=request_path, data=post_data)
            self.assertIn(b'<div class="success"></div>', response.content)
            self.assertEqual(response.context[2].template_name, "djangocms_transfer/placeholder_close_frame.html")

            structure_data = json.loads(response.context["structure_data"])
            # Since the import file is empty, only child plugins gets
            # passed to the frontend data bridge
            self.assertEqual(len(structure_data["plugins"]), 1)

            # POST placeholder import: simple data
            post_data["import_file"] = SimpleUploadedFile(
                "file.txt",
                bytes(json.dumps(self._get_expected_placeholder_export_data()), "utf-8")
            )
            response = self.client.post(path=request_path, data=post_data)
            self.assertIn(b'<div class="success"></div>', response.content)

            structure_data = json.loads(response.context["structure_data"])
            # newly imported plugin and child plugin gets passed to the frontend data bridge
            self.assertEqual(len(structure_data["plugins"]), 2)

    def test_import_plugin_views_for_plugin(self):
        # create plugins
        text_plugin = self._create_plugin()
        pluginimporter_plugin = self._create_plugin("PluginImporter")
        self._create_plugin(plugin_type="LinkPlugin", parent=text_plugin)

        with self.login_user_context(self.user):
            # GET plugin import
            response = self.client.get(
                admin_reverse("cms_import_plugins") + f"?language=en&plugin={text_plugin.pk}"
            )
            self.assertEqual(response.templates[0].name, "djangocms_transfer/import_plugins.html")
            self.assertEqual(response.context["form"].initial["plugin"], text_plugin.cmsplugin_ptr)

            # empty POST plugin import
            request_path = admin_reverse("cms_import_plugins") + f"?language=en&plugin={text_plugin.pk}"
            post_data = {
                "language": "en", "plugin": [text_plugin.id],
                "import_file": SimpleUploadedFile("file.txt", b"[null, null]")
            }
            with self.assertRaises(IndexError):
                self.client.post(path=request_path, data=post_data)

            # plugin import on TextPlugin
            post_data["import_file"] = SimpleUploadedFile(
                "file.txt",
                bytes(json.dumps(self._get_expected_placeholder_export_data()), "utf-8")
            )
            response = self.client.post(path=request_path, data=post_data)
            self.assertIn(b'<div class="success"></div>', response.content)

            # plugin import on PluginImporterPlugin
            request_path = admin_reverse("cms_import_plugins") + f"?language=en&plugin={pluginimporter_plugin.pk}"
            post_data = {"language": "en", "plugin": [pluginimporter_plugin.id]}
            post_data["import_file"] = SimpleUploadedFile(
                "file.txt",
                bytes(json.dumps(self._get_expected_placeholder_export_data()), "utf-8")
            )
            response = self.client.post(path=request_path, data=post_data)
            self.assertIn(b'<div class="success"></div>', response.content)

    def test_export_plugin_views(self):
        with self.login_user_context(self.user):
            response = self.client.get(admin_reverse("cms_export_plugins"))
            self.assertEqual(response.content, b"Form received unexpected values.")

        del self.user # self.get_request() checks for "user" attribute
        self.assertRaises(
            PermissionDenied,
            self.plugin_importer.export_plugins_view,
            self.get_request()
        )

    def test_export_plugin_views_for_plugin(self):
        # create plugins
        text_plugin = self._create_plugin()
        self._create_plugin(plugin_type="LinkPlugin", parent=text_plugin)

        with self.login_user_context(self.user):
            response = self.client.get(
                admin_reverse("cms_export_plugins") + f"?language=en&plugin={text_plugin.pk}"
            )
            self.assertEqual(response.status_code, 200)
            exported_data = json.loads(response.content)
            # exported content should be TextPlugin and its child (LinkPlugin).
            self.assertEqual(len(exported_data), 2)
            self.assertEqual(list(map(lambda i: i["plugin_type"], exported_data)), ["TextPlugin", "LinkPlugin"])

    def test_export_plugin_views_for_placeholder(self):
        placeholder = self.page_content.get_placeholders().get(slot="content")
        # create plugins in the placeholder
        text_plugin = self._create_plugin()
        self._add_plugin_to_page("PluginImporter")
        self._create_plugin(plugin_type="LinkPlugin", parent=text_plugin)

        with self.login_user_context(self.user):
            response = self.client.get(
                admin_reverse("cms_export_plugins") + f"?language=en&placeholder={placeholder.pk}"
            )
            self.assertEqual(response.status_code, 200)
            exported_data = json.loads(response.content)
            self.assertEqual(len(exported_data), 3)
            self.assertEqual(list(map(lambda i: i["plugin_type"], exported_data)), ["TextPlugin", "LinkPlugin", "PluginImporter"])

    def test_export_plugin_views_for_page(self):
        with self.login_user_context(self.user):
            response = self.client.get(
                admin_reverse("cms_export_plugins") + f"?language=en&cms_pagecontent={self.page_content.pk}"
            )
            self.assertEqual(response.status_code, 200)
            exported_data = json.loads(response.content)
            self.assertEqual(exported_data[0]["placeholder"], "content")
            self.assertEqual(len(exported_data[0]["plugins"]), 0)



__version__ = "2.0.1"

default_app_config = "djangocms_transfer.apps.TranferConfig"


def get_serializer_name(default="python"):
    from django.conf import settings

    return getattr(settings, "DJANGO_CMS_TRANSFER_SERIALIZER", default)


def custom_process_hook(transfer_hook, plugin, plugin_data):
    from importlib import import_module

    from django.conf import settings

    hook = getattr(settings, transfer_hook, None)
    if not hook:
        return plugin_data

    module, function = hook.rsplit(".", 1)
    try:
        func = getattr(import_module(module), function)
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Could not import '{hook}': {e}")

    return func(plugin, plugin_data)



import json
import os
import unittest
from tempfile import mkdtemp

from django.core.files import File

from djangocms_transfer.forms import PluginExportForm, PluginImportForm

from .abstract import FunctionalityBaseTestCase


class PluginExportFormTest(FunctionalityBaseTestCase):
    def test_get_filename(self):
        placeholder = self.page_content.get_placeholders().get(slot="content")
        plugin = self._create_plugin()

        data = {
            "plugin": plugin,
            "placeholder": placeholder,
            "cms_pagecontent": self.page_content,
            "language": "en",
        }

        with self.subTest("filename from page"):
            form = PluginExportForm(data=data)
            form.clean()
            self.assertEqual("home.json", form.get_filename())

        with self.subTest("filename from placeholder"):
            data["cms_pagecontent"] = None
            form = PluginExportForm(data=data)
            form.clean()
            self.assertEqual("home_content.json", form.get_filename())

        with self.subTest("filename from plugin"):
            data["placeholder"] = None
            form = PluginExportForm(data=data)
            form.clean()
            self.assertEqual("home_hello-world.json", form.get_filename())

        with self.subTest("filename from fallback"):
            data["plugin"] = None
            form = PluginExportForm(data=data)
            form.clean()
            self.assertEqual("plugins.json", form.get_filename())

    def test_validation(self):
        placeholder = self.page_content.get_placeholders().get(slot="content")
        plugin = self._create_plugin()

        with self.subTest("language missing"):
            form = PluginExportForm(data={})
            self.assertEqual(["This field is required."], form.errors["language"])

        with self.subTest("one of plugin/placeholder/page required"):
            form = PluginExportForm(data={"language": "en"})
            self.assertEqual(["A plugin, placeholder or page is required."], form.errors["__all__"])

        with self.subTest("cms_pagecontent + plugin given"):
            form = PluginExportForm(data={"language": "en", "cms_pagecontent": self.page_content, "plugin": plugin})
            self.assertEqual(
                ["Plugins can be imported to pages, plugins or placeholders. Not all three."],
                form.errors["__all__"],
            )

        with self.subTest("cms_pagecontent + placeholder given"):
            form = PluginExportForm(
                data={"language": "en", "cms_pagecontent": self.page_content, "placeholder": placeholder}
            )
            self.assertEqual(
                ["Plugins can be imported to pages, plugins or placeholders. Not all three."],
                form.errors["__all__"],
            )

        with self.subTest("plugin + placeholder given"):
            form = PluginExportForm(data={"language": "en", "plugin": plugin, "placeholder": placeholder})
            self.assertEqual(
                ["Plugins can be imported to pages, plugins or placeholders. Not all three."],
                form.errors["__all__"],
            )

    def test_run_export(self):
        placeholder = self.page_content.get_placeholders().get(slot="content")
        plugin = self._create_plugin()

        data = {
            "plugin": plugin,
            "placeholder": None,
            "cms_page": None,
            "language": "en",
        }

        with self.subTest("export plugin"):
            form = PluginExportForm(data=data)
            form.clean()
            actual = json.loads(form.run_export())
            self.assertEqual(self._get_expected_plugin_export_data(), actual)

        with self.subTest("export placeholder"):
            data["placeholder"] = placeholder
            data["plugin"] = None
            form = PluginExportForm(data=data)
            form.clean()
            actual = json.loads(form.run_export())
            self.assertEqual(self._get_expected_placeholder_export_data(), actual)

        with self.subTest("export page"):
            data["cms_pagecontent"] = self.page_content
            data["placeholder"] = None
            form = PluginExportForm(data=data)
            form.clean()
            actual = json.loads(form.run_export())
            self.assertEqual(self._get_expected_page_export_data(), actual)


class PluginImportFormTest(FunctionalityBaseTestCase):
    def test_validation(self):
        page = self.page
        placeholder = self.page_content.get_placeholders().get(slot="content")
        plugin = self._create_plugin()
        file_ = self._get_file()

        with self.subTest("file missing"):
            form = PluginImportForm(data={})
            self.assertEqual(["This field is required."], form.errors["import_file"])

        with self.subTest("language missing"):
            form = PluginImportForm(data={"import_file": file_})
            self.assertEqual(["This field is required."], form.errors["language"])

        # TODO: when setting the form, the `form.errors` is filled for "missing
        # import_file" although it is given/set in `data`
        self.skipTest("TODO: fix validation with 'import_file'")

        with self.subTest("one of plugin/placeholder/page required"):
            form = PluginImportForm(data={"language": "en", "import_file": file_})
            self.assertEqual(["A plugin, placeholder or page is required."], form.errors["__all__"])

        with self.subTest("cms_page + plugin given"):
            form = PluginImportForm(data={"import_file": file_, "language": "en", "cms_page": page, "plugin": plugin})
            self.assertEqual(
                ["Plugins can be imported to pages, plugins or placeholders. Not all three."],
                form.errors["__all__"],
            )

        with self.subTest("cms_page + placeholder given"):
            form = PluginImportForm(
                data={"import_file": file_, "language": "en", "cms_page": page, "placeholder": placeholder},
            )
            self.assertEqual(
                ["Plugins can be imported to pages, plugins or placeholders. Not all three."],
                form.errors["__all__"],
            )

        with self.subTest("plugin + placeholder given"):
            form = PluginImportForm(
                data={"import_file": file_, "language": "en", "plugin": plugin, "placeholder": placeholder},
            )
            self.assertEqual(
                ["Plugins can be imported to pages, plugins or placeholders. Not all three."],
                form.errors["__all__"],
            )

    @unittest.skip("TODO: fix validation with 'import_file'")
    def test_run_import(self):
        # TODO: when setting the form, the `form.errors` is filled for "missing
        # import_file" although it is given/set in `data`
        page = self.page
        placeholder = self.page_content.get_placeholders().get(slot="content")
        plugin = self._create_plugin()
        file_ = self._get_file()

        data = {
            "plugin": plugin,
            "placeholder": None,
            "cms_page": None,
            "language": "en",
            "import_file": file_,
        }

        with self.subTest("import plugin"):
            form = PluginImportForm(data=data)
            form.clean()
            form.run_import()

        with self.subTest("import placeholder"):
            data["placeholder"] = placeholder
            data["plugin"] = None
            form = PluginImportForm(data=data)
            form.clean()
            form.run_import()

        with self.subTest("import page"):
            data["cms_page"] = page
            data["placeholder"] = None
            form = PluginImportForm(data=data)
            form.clean()
            form.run_import()

    def _get_file(self):
        content = json.dumps(self._get_expected_plugin_export_data())

        tmp_dir = mkdtemp()
        filename = os.path.join(tmp_dir, "dummy-test.json")
        with open(filename, "w") as f:
            f.write(content)

        return File(open(filename, "rb"), name=filename)



import json

from djangocms_transfer.exporter import (
    export_page,
    export_placeholder,
    export_plugin,
)

from .abstract import FunctionalityBaseTestCase


class ExportTest(FunctionalityBaseTestCase):
    def test_export_plugin(self):
        plugin = self._create_plugin()

        actual = json.loads(export_plugin(plugin))

        self.assertEqual(self._get_expected_plugin_export_data(), actual)

    def test_export_placeholder(self):
        placeholder = self.page_content.get_placeholders().get(slot="content")

        with self.subTest("empty placeholder"):
            actual = json.loads(export_placeholder(placeholder, "en"))
            self.assertEqual([], actual)

        with self.subTest("placeholder with plugin"):
            self._create_plugin()
            actual = json.loads(export_placeholder(placeholder, "en"))
            self.assertEqual(self._get_expected_placeholder_export_data(), actual)

    def test_export_page(self):
        with self.subTest("empty page"):
            actual = json.loads(export_page(self.page_content, "en"))
            self.assertEqual([{"placeholder": "content", "plugins": []}], actual)

        with self.subTest("page with plugin"):
            self._create_plugin()
            actual = json.loads(export_page(self.page_content, "en"))
            self.assertEqual(self._get_expected_page_export_data(), actual)



import json

from cms.models import CMSPlugin
from djangocms_text.utils import plugin_to_tag

from djangocms_transfer.datastructures import (
    ArchivedPlaceholder,
    ArchivedPlugin,
)
from djangocms_transfer.exporter import export_placeholder, export_plugin
from djangocms_transfer.importer import import_plugins, import_plugins_to_page

from .abstract import FunctionalityBaseTestCase


class ImportTest(FunctionalityBaseTestCase):
    def test_import(self):
        pagecontent = self.page_content
        placeholder = pagecontent.get_placeholders().get(slot="content")
        plugin = self._create_plugin()

        # create link plugin
        link_plugin = self._create_plugin(plugin_type="LinkPlugin", parent=plugin)

        # Add plugin to text body
        plugin.body = f"{plugin.body} {plugin_to_tag(link_plugin)}"
        plugin.save()

        link_plugin_data = ArchivedPlugin(
            **json.loads(export_plugin(link_plugin))[0]
        )
        plugin_data = ArchivedPlugin(**json.loads(export_plugin(plugin))[0])
        placeholder_data = ArchivedPlaceholder(
            "content",
            [ArchivedPlugin(**data) for data in json.loads(export_placeholder(placeholder, "en"))],
        )

        with self.subTest("import plugins"):
            import_plugins([plugin_data, link_plugin_data], placeholder, "en")

            # test import updates child plugin
            new_plugin, new_link_plugin = map(
                lambda plugin: plugin.get_bound_plugin(), CMSPlugin.objects.filter(pk__in=[3,4])
            )
            self.assertEqual(
                new_plugin.body,
                f"{self.TEXT_BODY} {plugin_to_tag(new_link_plugin)}"
            )

        with self.subTest("import placeholder"):
            import_plugins(placeholder_data.plugins, placeholder, "en")

        with self.subTest("import page"):
            import_plugins_to_page([placeholder_data], pagecontent, "en")
